@@ -1,9 +1,39 @@
-import re
+"""
+A wrapper for the webservice (RPC API) for FusionDirectory.
+"""
+
 import requests
 
 
 class FusionDirectoryAPI:
-    def __init__(self, host, user, password, database, login=True):
+    def __init__(
+        self,
+        host,
+        user,
+        password,
+        database,
+        verify_cert=True,
+        login=True,
+        enforce_encryption=True,
+        client_id="python_api_wrapper",
+    ):
+        """
+        Log in to FusionDirectory server (Request a session ID)
+
+        Args:
+            host: The address of the FusionDirectory host including protocol (https://)
+            user: The name of the FD user to log in as
+            password: The password of the FD user
+            database: The database to use (As seen in FD GUI)
+            verify_cert: Verify server certificate (Default: True).
+            See requests documentation for options (https://2.python-requests.org/en/master/user/advanced/#verification)
+            login: Automatically log in on object instantiation (Default: True)
+            enforce_encryption: Raise an exception if traffic is unencrypted (Not https:// in host)
+        """
+
+        # Must encrypt traffic
+        if "https://" not in host and enforce_encryption:
+            raise ValueError("Unencrypted host not allowed: {host}")
 
         # The session to use for all requests
         self._session = requests.Session()
@@ -14,8 +44,11 @@ class FusionDirectoryAPI:
         # Log in to get this ID from FD
         self._session_id = None
 
-        # Send this ID wuth all requests
-        self._id = "API"
+        # Pass to requests
+        self._verify_cert = verify_cert
+
+        # Send this ID with all requests
+        self._client_id = client_id
 
         # Login to FD (Get a session_id)
         if login:
@@ -39,7 +72,7 @@ class FusionDirectoryAPI:
         r = self._post(data)
         # Api returns nothing on success, so anything is an error
         if r:
-            raise Exception(r)
+            raise LookupError(r)
         else:
             return True
 
@@ -53,7 +86,8 @@ class FusionDirectoryAPI:
 
     def get_fields(self, object_type, object_dn=None, tab=None):
         """
-        Get all fields of an object type as they are stored in FusionDirectory
+        Get all fields of an object type as they are stored in FusionDirectory.
+        Not very usefull unless using data for a GUI.
 
         Args:
             object_type (str): The type of object to get the fields for
@@ -69,7 +103,7 @@ class FusionDirectoryAPI:
         }
         return self._post(data)
 
-    def get_number_of_objects(self, object_type, ou=None, filter=""):
+    def get_number_of_objects(self, object_type, ou=None, filter=None):
         """
         Get the number of a given object type limited by OU and/or filter
 
@@ -80,24 +114,80 @@ class FusionDirectoryAPI:
 
         Returns:
             The number of objects of type object_type in the OU (int)
+            Some object types including "DASHBOARD", "SPECIAL", "LDAPMANAGER"
+            maybe others, results in a None value from the API. This functions
+            returns -1 in those cases.
         """
+
         data = {
             "method": "count",
             "params": [self._session_id, object_type, ou, filter],
         }
-        return self._post(data)
+        r = self._post(data)
+        # The API returns None for some object types
+        # I'm aware of these: ["DASHBOARD", "SPECIAL", "LDAPMANAGER"]
+        # Let's return -1, so we always return an int
+        if r == None:
+            r = -1
+        # assert type(r) == int
+        return r
 
-    def get_id(self):
+    def get_session_id(self):
         """
         Get current session ID
 
         Returns:
             The currents session id (str)
         """
+        # Not logged in
+        if not self._session_id:
+            return self._session_id
+
         data = {"method": "getId", "params": [self._session_id]}
         return self._post(data)
 
-    def get_objects(self, object_type, attributes=None, ou=None, filter=""):
+    def get_object(self, object_type, dn, attributes={"objectClass": "*"}):
+        """
+        Get attributes for a single object.
+
+        Arguments:
+            object_type (str): The object type to list
+            dn: The DN of the object to retrieve
+            attributes: The attributes to fetch.
+            If this is a single value, the resulting dictionary will have
+            for each dn the value of this attribute.
+            If this is an array, the keys must be the wanted attributes,
+            and the values can be either 1, '*', 'b64' or 'raw'
+            depending if you want a single value or an array of values.
+            Other values are considered to be 1.
+            'raw' means untouched LDAP value and is only useful for dns.
+            'b64' means an array of base64 encoded values and is mainly useful for binary attributes.
+
+        Returns:
+            A dictionary of attributes for the object
+        """
+        # Grab the left most part of the dn (uid=??) as filter
+        f = dn.split(",")[0]
+        filter = f"({f})"
+        # DN with out left most part is OU (Base for search)
+        ou = ",".join(dn.split(",")[1:])
+        data = {
+            "method": "ls",
+            "params": [self._session_id, object_type, attributes, ou, filter],
+        }
+        # FIXME: Check what data is returned if no objects are found
+        r = self._post(data)
+        # API returns an empty list is on no results. I need a dict.
+        if r == []:
+            r = {}
+        else:
+            # Api returns the user's data as value for key DN. We just
+            # want the value (The LDAP fields)
+            r = r[dn]
+        # assert type(r) == dict
+        return r
+
+    def get_objects(self, object_type, attributes=None, ou=None, filter=None):
         """
         Get objects of a certain type. Potentially with LDAP attributes and limited
         by OU and/or a filter.
@@ -123,7 +213,13 @@ class FusionDirectoryAPI:
             "method": "ls",
             "params": [self._session_id, object_type, attributes, ou, filter],
         }
-        return self._post(data)
+        # FIXME: Check what data is returned if no objects are found
+        r = self._post(data)
+        # An empty list is returned on no results. I need a dict.
+        if r == []:
+            r = {}
+        # assert type(r) == dict
+        return r
 
     def get_databases(self):
         """
@@ -131,7 +227,8 @@ class FusionDirectoryAPI:
         values for the 'database' argument in login()
 
         Returns:
-            A list of databases managed by FusionDirectory
+            A dict of databases managed by FusionDirectory. Key is id,
+            value is displayable name.
         """
         data = {"method": "listLdaps", "params": []}
         return self._post(data)
@@ -167,7 +264,7 @@ class FusionDirectoryAPI:
         }
         return self._post(data)
 
-    def get_info(self, object_type):
+    def get_object_type_info(self, object_type):
         """
         Get the information on an object type
 
@@ -180,19 +277,25 @@ class FusionDirectoryAPI:
         data = {"method": "infos", "params": [self._session_id, object_type]}
         return self._post(data)
 
-    def is_user_locked(self, user_dn):
+    def user_is_locked(self, user_dn):
         """
-        Get the lock state of one, or more, users
+        Is the user locked?
 
         Args:
             user_dn (str/list): A single DN or a list of DNs
 
         Returns:
-            result: A dictionary with user DNs as keys and the
-            lock state as values: 0 = unlocked, 1 = locked
+            True if user locked. False if not locked.
         """
+        # API accepts both list of DNs and a single DN. I don't
+        if type(user_dn) != str:
+            raise ValueError("user_dn must be a string")
         data = {"method": "isUserLocked", "params": [self._session_id, user_dn]}
-        return self._post(data)
+        r = self._post(data)
+        # API returns a dict with DN as key, and 0 or 1 in value
+        # assert len(r) == 1
+        # Return value in dict as bool
+        return bool(list(r.values())[0])
 
     def lock_user(self, user_dn):
         """
@@ -248,6 +351,8 @@ class FusionDirectoryAPI:
         """
         data = {"method": "recoveryGenToken", "params": [self._session_id, email]}
         r = self._post(data)
+        # FIXME: I get no UID in the dict (Value == None).
+        # According to the documentation, I should?
         return r["token"]
 
     def get_template(self, object_type, template_dn):
@@ -400,11 +505,11 @@ class FusionDirectoryAPI:
         Returns:
             result: The value of the key 'result' in the JSON returned by the server
         """
-        # All requests must have this, non session, ID (Don't know why?)
-        data["id"] = self._id
+        # Client ID (Se we can identify calls in server logs?
+        data["id"] = self._client_id
 
         # Post
-        r = self._session.post(self._url, json=data)
+        r = self._session.post(self._url, json=data, verify=self._verify_cert)
 
         # Raise exception on error codes
         r.raise_for_status()
@@ -413,10 +518,10 @@ class FusionDirectoryAPI:
         r = r.json()
 
         if r["error"]:
-            raise Exception(f"FD returned error: {r['error']}")
+            raise LookupError(f"FD returned error: {r['error']}")
         else:
             # The result value can have the key errors with a list
             if type(r["result"]) == dict and r["result"].get("errors"):
-                raise Exception("".join(r["result"]["errors"]))
+                raise LookupError("".join(r["result"]["errors"]))
             else:
                 return r["result"]
